from argparse import ArgumentParser, BooleanOptionalAction, Namespace

from .. import cfg, get_client
from ..room import Room, RoomId, RoomMembership, RoomType, StateEvent
from .base import Command


class RoomCommand(Command):
    command = 'room'
    help = 'Room management'

    def add_arguments(self, parser: ArgumentParser) -> None:
        subparsers = parser.add_subparsers(metavar='SUBCOMMAND', required=True)

        CreateRoomCommand.add_as_subparser(self, subparsers)
        DestroyRoomCommand.add_as_subparser(self, subparsers)


class CreateRoomCommand(Command):
    command = 'create'
    help = 'Create a new room'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            '--space', '-s',
            action='store_true',
            help='Create the new room as a space.',
        )
        parser.add_argument(
            '--encrypted',
            action=BooleanOptionalAction,
            help='Enable encryption for the new room',
        )
        parser.add_argument(
            '--category',
            default='default',
            help='The room category, that decides it\'s properites',
        )
        parser.add_argument(
            '--private',
            action=BooleanOptionalAction,
            default=False,
            help="Don't allow space members to join this room",
        )
        parser.add_argument(
            'displayname',
            help='Display name of the new room',
        )

    def handle(self, args: Namespace) -> None:
        space = RoomId(cfg.get(
            f'room_options::{args.category}',
            'space',
            fallback=cfg.get('room_options', 'space'),
        ))

        initial_state: list[StateEvent] = []
        if not args.private and space:
            initial_state.append({
                'type': 'm.room.join_rules',
                'state_key': '',
                'content': {
                    'join_rule': 'restricted',
                    'allow': [
                        {
                            'type': 'm.room_membership',
                            'room_id': space,
                        }
                    ]
                }
            })

        room = Room.create(
            args.category,
            args.displayname,
            type=RoomType.SPACE if args.space else RoomType.ROOM,
            initial_state=initial_state,
            encrypted=args.encrypted,
        )
        if space:
            room.add_to_space(space, mutual=not args.private)

        print(f'Create new room with room id: "{room.room_id}".')


class DestroyRoomCommand(Command):
    command = 'destroy'
    help = 'Destroy a room'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            'room_id',
            help='The room id of the room to delete',
        )

    def handle(self, args: Namespace) -> None:
        room = Room(args.room_id, '__delete__')

        bot = get_client().whoami()

        for user in room.members():
            if user != bot:
                room.set_membership(user, RoomMembership.LEAVE)

        room.set_membership(bot, RoomMembership.LEAVE)
