from argparse import Namespace
from configparser import SectionProxy

from .. import cfg, get_client
from ..power import get_power_levels
from ..provider import Office, User
from ..room import HistoryVisibility, Room, RoomId, RoomMembership
from .base import Command

OFFICE_CATEGORY = '__office__'


class SyncCommand(Command):
    command = 'sync'
    help = 'Sync state'

    def _sync_power_levels(self, room: Room) -> None:
        desired = get_power_levels(room.category)
        current = room.get_power_levels()

        if desired != current:
            room.set_power_levels(desired)

    def _sync_members(self, room: Room, room_cfg: SectionProxy) -> None:
        current = set(room.members())

        groups = set(room_cfg.get('groups', fallback='').split())
        users = set(map(lambda u: u.mxid, filter(
            lambda u: set(u.groups) & groups,
            User.all(),
        )))
        users.add(get_client().whoami())

        if not room_cfg.getboolean('public', fallback=False):
            groups_allow = set(
                room_cfg.get('groups_allow', fallback='').split()
            )
            users_allow = set(map(lambda u: u.mxid, filter(
                lambda u: set(u.groups) & groups_allow,
                User.all(),
            )))

            # Remove
            for user in current - users - users_allow:
                room.set_membership(user, RoomMembership.LEAVE)

        # Add
        for user in users - current:
            room.set_membership(user, RoomMembership.INVITE)

    def handle(self, args: Namespace) -> None:
        for section_name in filter(
            lambda s: s.startswith('room::'),
            cfg.sections(),
        ):
            section = cfg[section_name]
            room = Room(
                RoomId(section.get('room-id')), section.get('category')
            )

            self._sync_power_levels(room)
            self._sync_members(room, section)

        # Offices
        for office in Office.all():
            if office.room is None:
                room = Room.create(
                    OFFICE_CATEGORY,
                    office.displayname,
                    history_visibility=HistoryVisibility.INVITED,
                )
                space = cfg.get(
                    f'room_options::{OFFICE_CATEGORY}',
                    'space',
                    fallback=cfg.get('room_options', 'space'),
                )
                if space:
                    room.add_to_space(RoomId(space))
                office.set_room(room.room_id)
            else:
                room = Room(RoomId(office.room), OFFICE_CATEGORY)

            self._sync_power_levels(room)

            current = set(room.members())
            desired = set(office.members)
            desired.add(get_client().whoami())

            # Remove
            for user in current - desired:
                room.set_membership(user, RoomMembership.LEAVE)

            # Add
            for user in desired - current:
                room.set_membership(user, RoomMembership.INVITE)
