from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, NewType, Optional, TypedDict, cast
from urllib.parse import quote

import requests
from requests.auth import AuthBase

if TYPE_CHECKING:
    from .power import PowerLevels

RoomId = NewType('RoomId', str)


class Error(Exception):
    pass


class ApiError(Error):
    """
    The api returned invalid data
    """
    pass


class StateEvent(TypedDict):
    type: str
    state_key: Optional[str]
    content: dict[str, Any]


class RoomVisibility(Enum):
    PUBLIC = 'public'
    PRIVATE = 'private'


class RoomPreset(Enum):
    PRIVATE_CHAT = 'private_chat'
    PUBLIC_CHAT = 'public_chat'
    TRUSTED_PRIVATE_CHAT = 'trusted_private_chat'


class RoomMembership(Enum):
    INVITE = 'invite'
    JOIN = 'join'
    KNOCK = 'knock'
    LEAVE = 'leave'
    BAN = 'ban'


class HTTPBearerAuth(AuthBase):
    """
    Attaches HTTP Bearer Token Authentication to the given Request object.
    """

    def __init__(self, token: Optional[str]):
        self.token = token

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.token == other.token

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __call__(
        self, r: requests.PreparedRequest
    ) -> requests.PreparedRequest:
        if self.token is not None:
            r.headers['Authorization'] = f'Bearer {self.token}'
        return r


class Client:
    compatible_versions = ['r0.6.1']
    token: Optional[str]

    def __init__(self, homeserver: str, device_id: Optional[str]):
        self.homeserver = homeserver
        self.device_id = device_id

        version = self.get_prefered_version()
        if version is None:
            raise ValueError("No common api version support with homeserver")

        self.version = version

    def _endpoint(self, endpoint: str) -> str:
        return f'{self.homeserver}/_matrix/client{endpoint}'

    def versions(self) -> Version:
        r = requests.get(self._endpoint('/versions'))
        r.raise_for_status()
        return Version.from_dict(r.json())

    def get_prefered_version(self) -> Optional[str]:
        available_versions = set(self.versions().versions)

        for version in self.compatible_versions:
            if version in available_versions:
                return version

        # No versions match
        return None

    def login(
        self,
        user: str,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ) -> None:
        assert password is not None or token is not None, (
            "At least one of password and token is required"
        )

        if token is not None:
            self.token = token
            return

        # Is not None because token is None
        password = cast(str, password)

        r = requests.post(self._endpoint('/r0/login'), json={
            'type': 'm.login.password',
            'identifier': {
                'type': 'm.id.user',
                'user': user,
            },
            'password': password,
        })
        r.raise_for_status()

        data = r.json()
        access_token = data['access_token']
        if not isinstance(access_token, str):
            raise ApiError("access_token must be a string")
        self.token = access_token

    def whoami(self) -> str:
        r = requests.get(
            self._endpoint('/r0/account/whoami'),
            auth=HTTPBearerAuth(self.token),
        )
        r.raise_for_status()
        user_id = r.json()['user_id']
        if not isinstance(user_id, str):
            raise ApiError('user_id must be a string')
        return user_id

    def create_room(
        self,
        visibility: Optional[RoomVisibility] = None,
        room_alias_name: Optional[str] = None,
        name: Optional[str] = None,
        topic: Optional[str] = None,
        invite: Optional[list[str]] = None,
        room_version: Optional[str] = None,
        creation_content: Optional[dict[str, Any]] = None,
        initial_state: Optional[list[StateEvent]] = None,
        preset: Optional[RoomPreset] = None,
        is_direct: Optional[bool] = None,
        power_level_content_override: Optional[PowerLevels] = None,

        # Helper
        enable_encryption: bool = False,
    ) -> RoomId:
        data = {
            'visibility': visibility and visibility.value,
            'room_alias_name': room_alias_name,
            'name': name,
            'topic': topic,
            'invite': invite,
            'room_version': room_version,
            'creation_content': creation_content,
            'initial_state': initial_state,
            'preset': preset and preset.value,
            'is_direct': is_direct,
            'power_level_content_override': power_level_content_override,
        }
        r = requests.post(
            self._endpoint('/r0/createRoom'),
            auth=HTTPBearerAuth(self.token),
            json={k: v for k, v in data.items() if v is not None},
        )
        r.raise_for_status()
        room_id = r.json()['room_id']
        if not isinstance(room_id, str):
            raise ApiError("Expected room_id to be a string")
        return RoomId(room_id)

    def room_members(
        self,
        room_id: RoomId,
    ) -> Iterator[MemberEvent]:
        r = requests.get(
            self._endpoint(f'/r0/rooms/{room_id}/members'),
            auth=HTTPBearerAuth(self.token),
            json={'not_membership': 'leave'},
        )
        r.raise_for_status()
        data = r.json()['chunk']
        if not isinstance(data, list):
            raise ApiError('Expected chunk to be a list')
        for event in data:
            yield MemberEvent.from_dict(event)

    def room_get_state_event(
        self,
        room_id: RoomId,
        event_type: str,
        state_key: str,
    ) -> Optional[dict[str, Any]]:
        r = requests.get(
            self._endpoint(
                f'/r0/rooms/{room_id}/state/{event_type}/{quote(state_key)}'
            ),
            auth=HTTPBearerAuth(self.token),
        )
        if r.status_code == 404:
            return None

        r.raise_for_status()
        return cast(dict[str, Any], r.json())

    def room_send_state_event(
        self,
        room_id: RoomId,
        event_type: str,
        state_key: str,
        content: dict[str, Any],
    ) -> None:
        q_room_id = quote(room_id)
        q_event_type = quote(event_type)
        q_state_key = quote(state_key)
        r = requests.put(
            self._endpoint(
                f'/r0/rooms/{q_room_id}/state/{q_event_type}/{q_state_key}'
            ),
            auth=HTTPBearerAuth(self.token),
            json=content,
        )
        r.raise_for_status()


@dataclass
class Version:
    versions: list[str]
    unstable_features: Optional[dict[str, bool]]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Version:
        versions = data['versions']
        if not (
            isinstance(versions, list)
            and all(map(lambda e: isinstance(e, str), versions))
        ):
            raise ApiError("data['versions'] must be a list of strings")

        unstable_features = data.get('unstable_features')
        if unstable_features is not None:
            if not (
                isinstance(unstable_features, dict) and all(map(
                    lambda k: isinstance(k, str),
                    unstable_features.keys()
                )) and all(map(
                    lambda v: isinstance(v, bool),
                    unstable_features.values()
                ))
            ):
                raise ApiError(
                    "data['unstable_features'] must be a mapping"
                    " from strings to booleans"
                )

        return cls(versions, unstable_features)


@dataclass
class MemberEvent:
    state_key: str
    content: MemberEventContent

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> MemberEvent:
        state_key = data['state_key']
        if not isinstance(state_key, str):
            raise ApiError("data['state_key'] must be a string")

        content = data['content']
        if not isinstance(content, dict):
            raise ApiError("data['content'] must be a mapping")

        return cls(state_key, MemberEventContent.from_dict(content))


@dataclass
class MemberEventContent:
    membership: RoomMembership

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> MemberEventContent:
        return cls(RoomMembership(data['membership']))
