from collections.abc import Iterable
from configparser import SectionProxy
from typing import TypedDict

from . import cfg, get_client
from .provider import User

_power_levels_users = {}

PowerLevelsEvents = dict[str, int]
PowerLevelsUsers = dict[str, int]


class PowerLevelsNotifications(TypedDict, total=False):
    room: int


class PowerLevels(TypedDict, total=False):
    ban: int
    events: PowerLevelsEvents
    events_default: int
    historical: int
    invite: int
    kick: int
    redact: int
    state_default: int
    users: PowerLevelsUsers
    users_default: int
    notifications: PowerLevelsNotifications


def get_power_sections(room_category: str) -> Iterable[SectionProxy]:
    defaults = cfg['power']
    specific_section = f'power::{room_category}'

    if specific_section in cfg:
        return (defaults, cfg[specific_section])
    else:
        return (defaults,)


def get_group_power(room_category: str) -> dict[str, int]:
    group_power = {}

    for power in get_power_sections(room_category):
        for g in power:
            if g.startswith('@'):  # Ignore mxids
                continue

            group_power[g] = power.getint(g)

    return group_power


def get_power_override(room_category: str) -> dict[str, int]:
    mxid_power = {}

    for power in get_power_sections(room_category):
        for g in power:
            if not g.startswith('@'):  # Ignore non mxids
                continue

            mxid_power[g] = power.getint(g)

    return mxid_power


def get_power_levels_users(
    room_category: str, force_refresh: bool = False
) -> PowerLevelsUsers:
    global _power_levels_users

    myself = get_client().whoami()

    if room_category not in _power_levels_users or force_refresh:
        group_power = get_group_power(room_category)

        power_levels = {}
        for user in User.all():
            # Get highest power level
            power = max(map(
                group_power.__getitem__,
                set(user.groups) & set(group_power),
            ), default=None)

            if power is not None:
                power_levels[user.mxid] = power

        power_override = get_power_override(room_category)
        for mxid, power in power_override.items():
            power_levels[mxid] = power

        power_levels[myself] = 1000
        _power_levels_users[room_category] = power_levels

    return _power_levels_users[room_category]


def get_power_levels_events(room_category: str) -> PowerLevelsEvents:
    events: dict[str, int] = {}

    defaults = cfg['power_levels']
    specific_section = f'power_levels::{room_category}'

    sections: Iterable[SectionProxy]
    if specific_section in cfg:
        sections = (defaults, cfg[specific_section])
    else:
        sections = (defaults,)

    for section in sections:
        for event in section:
            # Ignore special keys
            if event.startswith('_'):
                continue

            if not section[event]:
                # Explicit remove request
                del events[event]
            events[event] = section.getint(event)

    return events


def get_power_levels(
    room_category: str,
) -> PowerLevels:

    # function 'c' returns the config value from specific section and falls
    # back to default section.
    # It also automatically appends the underscore that is used for the special
    # settings
    default = cfg['power_levels']
    specific_section = f'power_levels::{room_category}'
    if specific_section not in cfg:
        def c(key: str) -> int:
            return default.getint(f'_{key}')
    else:
        def c(key: str) -> int:
            key = f'_{key}'
            return cfg[specific_section].getint(
                key, fallback=default.getint(key)
            )

    return {
        'users': get_power_levels_users(room_category),
        'events': get_power_levels_events(room_category),

        'users_default': c('users_default'),
        'events_default': c('events_default'),
        'state_default': c('state_default'),
        'ban': c('ban'),
        'kick': c('kick'),
        'invite': c('invite'),
        'redact': c('redact'),
        'historical': c('historical'),
    }
