"""
Provider to get Users awell as their groups and offices
"""
from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Any, Optional

import psycopg

from . import cfg

# TODO: use psycopg's query builder instead of blindly inseting the string
# into the query
MATRIX_DOMAIN = 'fsmi.uni-karlsruhe.de'

_db = None


def get_db() -> Any:
    global _db
    if _db is None:
        _db = psycopg.connect(cfg['provider::auth']['db'])
    return _db


class User:
    _db_id: int

    mxid: str
    groups: Sequence[str]

    def __repr__(self) -> str:
        return f"<User: {self.mxid}>"

    @classmethod
    def all(cls) -> Iterable[User]:
        db = get_db()
        with db.transaction(), db.cursor() as cur:
            cur.execute(
                f"""
                SELECT
                    benutzer.benutzer_id,
                    COALESCE(
                        benutzer.mxid,
                        '@' || benutzer.benutzername || ':{MATRIX_DOMAIN}'
                    ) AS mxid,
                    array_agg(rechte.rechtename) AS groups
                FROM benutzer
                JOIN acl.effektive_benutzer_rechte USING (benutzer_id)
                JOIN acl.rechte USING (rechteid)
                WHERE benutzer.benutzername IS NOT NULL
                GROUP BY benutzer.benutzer_id, benutzer.benutzername
                """,
            )
            for benutzer_id, mxid, groups in cur:
                user = cls()
                user._db_id = benutzer_id
                user.mxid = mxid
                user.groups = groups
                yield user


class Office:
    _db_id: int

    displayname: str
    members: Sequence[str]
    _room: Optional[str]

    def __repr__(self) -> str:
        return f"<Office: {self.displayname}>"

    @property
    def room(self) -> Optional[str]:
        return self._room

    def set_room(self, value: str) -> None:
        db = get_db()
        with db.transaction(), db.cursor() as cur:
            cur.execute(
                """
                UPDATE aemter
                SET matrix_room = %s
                WHERE amt_id = %s
                """,
                (value, self._db_id),
            )

        self._room = value

    @classmethod
    def all(cls) -> Iterable[Office]:
        db = get_db()
        with db.transaction(), db.cursor() as cur:
            cur.execute(
                f"""
                SELECT
                    aemter.amt_id,
                    CASE WHEN fachschaften.email_prefix IS NOT NULL
                        THEN fachschaften.name || ' '
                        ELSE ''
                    END || aemter_typen.name AS displayname,
                    array_agg(
                        COALESCE (
                            benutzer.mxid,
                            '@' || benutzer.benutzername || ':{MATRIX_DOMAIN}'
                        )
                    ) FILTER (
                        WHERE benutzer.benutzername IS NOT NULL
                    ) AS members,
                    matrix_room
                FROM aemter
                JOIN aemter_typen USING (amt_typ_id)
                JOIN fachschaften USING (fachschafts_id)
                JOIN aemter_und_benutzer USING (amt_id)
                JOIN benutzer USING (benutzer_id)
                GROUP BY
                    aemter.amt_id,
                    aemter_typen.name,
                    fachschaften.email_prefix,
                    fachschaften.name
                """,
            )

            for amt_id, displayname, members, matrix_room in cur:
                office = cls()
                office._db_id = amt_id
                office.displayname = displayname
                office.members = members
                office._room = matrix_room
                yield office
