from __future__ import annotations

from collections.abc import Iterator
from enum import Enum, auto
from typing import Any, Optional, cast

from . import cfg, get_client
from .matrix import RoomId, RoomMembership, RoomPreset, StateEvent
from .power import PowerLevels, get_power_levels

__all__ = [
    'RoomType', 'HistoryVisibility', 'Room',
]

# Reexports
__all__ += [
    'RoomId', 'RoomMembership', 'RoomPreset', 'StateEvent',
]


class RoomType(Enum):
    ROOM = auto()
    SPACE = auto()


class HistoryVisibility(Enum):
    WORLD_READABLE = 'world_readable'
    SHARED = 'shared'
    INVITED = 'invited'
    JOINED = 'joined'


class Room:
    def __init__(self, room_id: RoomId, category: str) -> None:
        self.room_id = room_id
        self.category = category

    def __repr__(self) -> str:
        return f'Room({self.room_id!r}, {self.category!r})'

    def members(self) -> Iterator[str]:
        for member_event in get_client().room_members(self.room_id):
            if member_event.content.membership in [
                RoomMembership.JOIN,
                RoomMembership.INVITE,
            ]:
                yield member_event.state_key

    def set_membership(self, user: str, membership: RoomMembership) -> None:
        get_client().room_send_state_event(
            self.room_id,
            'm.room.member',
            user,
            {'membership': membership.value},
        )

    def get_power_levels(self) -> PowerLevels:
        resp = get_client().room_get_state_event(
            self.room_id, 'm.room.power_levels', ''
        )
        # TODO: check response format
        return resp  # type: ignore

    def set_power_levels(self, power_levels: PowerLevels) -> None:
        get_client().room_send_state_event(
            self.room_id,
            'm.room.power_levels',
            '',
            cast(dict[str, Any], power_levels),
        )

    def add_to_space(self, space: RoomId, mutual: bool = True) -> None:
        # We only set our own server in via, because:
        #  - Most of the time the room was just created and is therefore empty
        #  - This bot user is intended to stay in the room
        via_server = get_client().whoami().partition(':')[2]

        get_client().room_send_state_event(
            self.room_id,
            'm.space.parent',
            space,
            {'via': [via_server], 'canonical': True},
        )

        if mutual:
            get_client().room_send_state_event(
                space,
                'm.space.child',
                self.room_id,
                {'via': [via_server]},
            )

    @classmethod
    def create(
        cls,
        category: str,
        name: Optional[str],
        type: RoomType = RoomType.ROOM,
        topic: Optional[str] = None,
        alias: Optional[str] = None,
        history_visibility: HistoryVisibility = HistoryVisibility.SHARED,
        encrypted: Optional[bool] = None,
        initial_state: Optional[list[StateEvent]] = None,
    ) -> Room:
        if initial_state is None:
            initial_state = []

        if encrypted or (encrypted is None and cfg.getboolean(
            f'room_options::{category}',
            'encrypted',
            fallback=cfg.getboolean('room_options', 'encrypted'),
        )):
            initial_state.append({
                'type': 'm.room.encryption',
                'state_key': '',
                'content': {
                    'algorithm': 'm.megolm.v1.aes-sha2',
                },
            })

        initial_state.append({
            'type': 'm.room.history_visibility',
            'state_key': '',
            'content': {
                'history_visibility': history_visibility.value,
            },
        })

        creation_content = {}
        if type == RoomType.SPACE:
            creation_content['type'] = 'm.space'

        room_id = get_client().create_room(
            room_alias_name=alias,
            name=name,
            topic=topic,
            creation_content=creation_content,
            preset=RoomPreset.PRIVATE_CHAT,
            power_level_content_override=get_power_levels(category),
            room_version='9',  # version 9 for restricted rooms
            initial_state=initial_state,
        )
        return cls(room_id, category)
